/*
 * Copyright (c) 2006
 *
 * written by Simon Barrick
 * simon@fourthwall.ndo.co.uk
 *
 * PolyCuspBevel SOP
 */

//#define BETA

//always include these
#include <UT/UT_DSOVersion.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <PRM/PRM_Include.h>
#include <UT/UT_Interrupt.h>

#include <GQ/GQ_Detail.h>
#include <GQ/GQ_PolyBevel.h>
#include <GB/GB_Group.h>
#include <GB/GB_Edge.h>
#include <GB/GB_PointRef.h>
#include <GB/GB_EdgeGroup.h>
#include <GEO/GEO_Closure.h>
#include <GU/GU_PrimGroup.h>
#include <SYS/SYS_Math.h>

#include "./SOP_PolyCuspBevel.H"


// This is were the code switches between the standard
// compile and the beta compile.
#ifdef BETA
static const char* const SOP_INTERNAL_NAME = "PolyCuspBevel_beta";
static const char* const SOP_UI_NAME = "Poly Cusp Bevel Beta";
#else
static const char* const SOP_INTERNAL_NAME = "PolyCuspBevel";
static const char* const SOP_UI_NAME = "Poly Cusp Bevel";
#endif


void
newSopOperator( OP_OperatorTable *table )
{
  table->addOperator
  (
    new OP_Operator
    (
      SOP_INTERNAL_NAME,             // Internal name
      SOP_UI_NAME,                   // UI name
      SOP_PolyCuspBevel::myConstructor,  //
      SOP_PolyCuspBevel::myTemplateList, //
      1,                             // Min Inputs
      1,                             // Max Inputs
      0                              // Local Variables
      //OP_FLAG_*                    // Additional Flag
    )

// Examples of additional flags include, OP_FLAG_GENERATOR, OP_FLAG_UNORDERED.
// See OP_Operator.h for more details.

  );
}

static PRM_Name typeMenuNames[] =
{
  PRM_Name( "flat", "Flat" ),

  PRM_Name( "low", "Low density round" ),

  PRM_Name( "high", "High density round" ),

  PRM_Name( "corner", "Corner" ),

  PRM_Name(0)

};

static PRM_ChoiceList   typeMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      typeMenuNames);

static PRM_Name methodMenuNames[] =
{
  PRM_Name( "rel", "Relative" ),

  PRM_Name( "abs", "Absolute" ),

  PRM_Name(0)

};

static PRM_ChoiceList   methodMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      methodMenuNames);

static PRM_Default PRMfortyfiveDefault(45);   // Default to 45
static PRM_Default PRMpointOneDefault(0.1f);   // Default to 0
static PRM_Default PRMGrpDefault(0.0f, "CuspGroup");


static PRM_Name names[] =
{
  PRM_Name( "ang", "Cusp Angle" ),

  PRM_Name( "type", "Type" ),

  PRM_Name( "inset", "Inset" ),

  PRM_Name( "method", "Method" ),

  PRM_Name( "reps", "Repetitions" ),

  PRM_Name( "norms", "Recompute normals" ),

  PRM_Name( "outgroup", "Create output group" ),

  PRM_Name( "groupname", "Group Name" ),

  PRM_Name( 0 )
};


PRM_Template
SOP_PolyCuspBevel::myTemplateList[] =
{
  PRM_Template(PRM_STRING,    1, &PRMgroupName, 0, &SOP_Node::primGroupMenu),

  PRM_Template(PRM_FLT_J, 1, &names[0], &PRMfortyfiveDefault, 0,
           &PRMnonNegativeRange),

  PRM_Template(PRM_ORD, 1, &names[1], PRMzeroDefaults, &typeMenu),

  PRM_Template(PRM_ORD, 1, &names[3], PRMoneDefaults, &methodMenu),

  PRM_Template(PRM_FLT_J, 1, &names[2], &PRMpointOneDefault),

  PRM_Template(PRM_INT_J, 1, &names[4], PRMzeroDefaults),

  PRM_Template(PRM_TOGGLE, 1, &names[5], PRMoneDefaults),

  PRM_Template(PRM_TOGGLE, 1, &names[6], PRMzeroDefaults),

  PRM_Template(PRM_STRING, 1, &names[7], &PRMGrpDefault),

  PRM_Template(),
};

const char *
SOP_PolyCuspBevel::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Geometry To Bevel";
     break;
   default:
     return "WTF?";
     break;
 }
}

//overriding constructor
OP_Node *
SOP_PolyCuspBevel::myConstructor( OP_Network *net, const char *name, OP_Operator *op )
{
  return new SOP_PolyCuspBevel( net, name, op );
}

SOP_PolyCuspBevel::SOP_PolyCuspBevel( OP_Network *net, const char *name, OP_Operator *op )
  : SOP_Node( net, name, op )
{
  setHighlight(1);
}

//destructor
SOP_PolyCuspBevel::~SOP_PolyCuspBevel() {}

unsigned
SOP_PolyCuspBevel::disableParms()
{
    unsigned changed = 0;

    changed += enableParm("groupname", GROUP());

    return changed;
}

OP_ERROR
SOP_PolyCuspBevel::cookInputGroups(OP_Context &context, int alone)
{
    // If we are called by the handle, then "alone" equals 1.  In that
    // case, we have to lock the inputs oursevles, and unlock them
    // before exiting this method.
    if (alone) if (lockInputs(context) >= UT_ERROR_ABORT) return error();

    UT_String  grp_name;

    // The "gdp" variable is only available if we are called from the SOP
    // itself.  So, if we are called by a handle, we have to get the
    // geometry oursevles.
    GU_Detail *pgdp = alone ? (GU_Detail *)inputGeo(0, context) : gdp;

    myGroup = 0;

    getGroups(grp_name);    // Get the group string.

    // If the group string is not null, then we try to parse the group.
    if (grp_name.isstring())
    {
      myGroup = parsePrimitiveGroups((const char *)grp_name, pgdp);

      // If the group is not valid, then the group string is invalid
      // as well.  Thus, we add an error to this SOP.
      if (!myGroup)
      {
          addError(SOP_ERR_BADGROUP, grp_name);
      }
    }
    else if (!alone)
    {
      // If no group string is specified, then we operate on the entire
      // geometry, so we highlight every point for this SOP.
      select(GU_SPoint);
    }

    // This is where we notify our handles (if any) if the inputs have changed.
    checkInputChanged(0, -1, myDetailGroupPair, pgdp, myGroup);

    // If we are called by the handles, then we have to unlock our inputs.
    if (alone)
    {
      destroyAdhocGroups();
      unlockInputs();
    }

    return error();
}

inline int
SOP_PolyCuspBevel::MOD(int a, int b)
{
  return (a>b?0:a);
}
//cook method

OP_ERROR
SOP_PolyCuspBevel::cookMySop( OP_Context &context )
{
  float now = context.myTime;

  // Before we do anything, we must lock our inputs.  Before returning,
  // we have to make sure that the inputs get unlocked.
  if( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  //copy input geomtry as we aren't creating any
  duplicateSource( 0, context );

  GQ_PolyBevel          polyBevel(gdp, METHOD(), INSET(now), TYPE(), REPS(now));
  GB_PrimitiveGroup     *edgeprimsgrp,*createdprimsgrp;
  GB_PointRefArray      prefarr(gdp,0);
  GEO_Primitive         *prim,*cprim;
  GEO_Point             *pt0,*pt1;
  GEO_Vertex            *vtx0,*vtx1;
  GB_EdgeGroup          edgeGroup(*gdp,"tmpedge");
  int                   i,test;
  GEO_Closure           primclosure(*gdp);

  float cuspang = SYScos(SYSdegToRad(ANG(now)))+1;

  edgeprimsgrp = createAdhocPrimGroup(*gdp, "tmpPrims");
  createdprimsgrp = createAdhocPrimGroup(*gdp, "createdPrims");

  // Here we determine which groups we have to work on.  We only
  //  handle prim groups.
  if (error() < UT_ERROR_ABORT && cookInputGroups(context) < UT_ERROR_ABORT)
  {
    GU_PrimGroup newGroup(gdp, createdprimsgrp);
    FOR_ALL_PRIMITIVES(gdp, prim)
      newGroup.group()->add(prim);

    FOR_ALL_OPT_GROUP_PRIMITIVES(gdp, myGroup, prim)
    {
      for (i = 0; i < prim->getVertexCount(); i++)
      {
        vtx0 = &prim->getVertex(i);
        pt0  = vtx0->getPt();
        vtx1 = &prim->getVertex(MOD(i+1,prim->getVertexCount()));
        pt1  = vtx1->getPt();
        primclosure.findPolysUsingEdge(*pt0, *pt1, edgeprimsgrp, prefarr);
        FOR_ALL_GROUP_PRIMITIVES(gdp, edgeprimsgrp, cprim)
        {
          //dot the prim normals prim and cprim
          float cusp = prim->computeNormal().dot(cprim->computeNormal())+1;
          if (cusp < cuspang)
            edgeGroup.add((GB_Element*)pt0,(GB_Element*)pt1,prim);
        }
      }
    }
    edgeprimsgrp = polyBevel.edgeBevel(edgeGroup);
    newGroup.boolean(GU_GROUP_BOOLOP_NONE,(GB_Group*)newGroup.group(),true,(GB_Group*)newGroup.group(),false);
    select(*(GB_BaseGroup *)newGroup.group(), 1);

    if (GROUP())
    {
      UT_String outgroupname;
      GROUPNAME(outgroupname);
      GU_PrimGroup outGroup(outgroupname, gdp);
      *outGroup.group() = *newGroup.group();

    }
  }

  destroyAdhocGroups(gdp);

  if (UPDATE())
  {
    test = gdp->findPointAttrib("N", GB_ATTRIB_VECTOR);
    if (test > -1)
        gdp->normal();
  }

  unlockInputs();

  return error();

}
