/*
 * Copyright (c) 2006
 *
 * written by Simon Barrick
 * simon@fourthwall.ndo.co.uk
 *
 * PolyCuspBevel SOP
 */

#ifndef __SOP_PolyCuspBevel_h__
#define __SOP_PolyCuspBevel_h__

#include <SOP/SOP_Node.h>

class SOP_PolyCuspBevel : public SOP_Node
{
public:
    static PRM_Template    myTemplateList[];
    static OP_Node    *myConstructor(OP_Network*, const char *,
                  OP_Operator *);

    // This method is created so that it can be called by handles.  It only
    // cooks the input group of this SOP.  The geometry in this group is
    // the only geometry manipulated by this SOP.
    virtual OP_ERROR     cookInputGroups(OP_Context &context, int alone = 0);


protected:
       SOP_PolyCuspBevel(OP_Network *net, const char *name, OP_Operator *op);
    virtual ~SOP_PolyCuspBevel();

    virtual unsigned      disableParms();
    virtual const char    *inputLabel(unsigned int) const;

    // Method to cook geometry for the SOP
    virtual OP_ERROR       cookMySop(OP_Context &context);

private:

  float ANG(float t)       { return evalFloat("ang", 0, t); }

  int   TYPE()             { return evalInt("type", 0, 0); }

  int   METHOD()           { return evalInt("method", 0, 0); }

  int   REPS(float t)      { return evalInt("reps", 0, t); }

  float INSET(float t)     { return evalFloat("inset", 0, t); }

  bool  UPDATE()           { return evalInt("norms", 0, 0); }

  int   MOD(int, int);

  int   GROUP()            { return evalInt("outgroup", 0, 0); }

  void  GROUPNAME(UT_String &str)  { evalString(str, "groupname", 0, 0); }


  void  getGroups(UT_String &str) { evalString(str, 0, 0, 0); }


  // This variable is used together with the call to the "checkInputChanged"
  // routine to notify the handles (if any) if the input has changed.
  GU_DetailGroupPair   myDetailGroupPair;

  // This is the group of geometry to be manipulated by this SOP and cooked
  // by the method "cookInputGroups".
  const GB_PrimitiveGroup *myGroup;

};

#endif
